/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ex: set tabstop=8 softtabstop=4 shiftwidth=4 expandtab: */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is developed for mozilla. The tables of paper sizes
 * and paper orientations are based on tables from nsPostScriptObj.h.
 *
 * The Initial Developer of the Original Code is
 * Kenneth Herron <kherron@newsguy.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifdef MOZ_LOGGING
#define FORCE_PR_LOG 1 /* Allow logging in the release build */
#endif /* MOZ_LOGGING */
#include "prlog.h"

#include "nsPaperPS.h"
#include "plstr.h"
#include "nsPSPrinters.h"
#include <math.h> 

#ifdef PR_LOGGING 
static PRLogModuleInfo *PaperSizePSLM = PR_NewLogModule("PaperSizePS");
#endif /* PR_LOGGING */
/* Macro to make lines shorter */
#define DO_PR_DEBUG_LOG(x) PR_LOG(PaperSizePSLM, PR_LOG_DEBUG, x)

#define MM_PER_PT (25.4 / 72.0)
#define HALF_INCH_PT 36.0
#define EPSILON 0.125

#define COUNTOF(x) (sizeof(x) / sizeof((x)[0]))

const nsPaperSizePS_ nsPaperSizePS::mList[] =
{
#define SIZE_MM(x)      (x)
#define SIZE_INCH(x)    ((x) * 25.4)
    { "A5",             SIZE_MM(148),   SIZE_MM(210),   PR_TRUE },
    { "A4",             SIZE_MM(210),   SIZE_MM(297),   PR_TRUE },
    { "A3",             SIZE_MM(297),   SIZE_MM(420),   PR_TRUE },
    { "Letter",         SIZE_INCH(8.5), SIZE_INCH(11),  PR_FALSE },
    { "Legal",          SIZE_INCH(8.5), SIZE_INCH(14),  PR_FALSE },
    { "Tabloid",        SIZE_INCH(11),  SIZE_INCH(17),  PR_FALSE },
    { "Executive",      SIZE_INCH(7.5), SIZE_INCH(10),  PR_FALSE },
#undef SIZE_INCH
#undef SIZE_MM
};

nsCUPSShim nsPaperSizeCUPS::mCups;

/* ~nsIPaperSizePS() is virtual, so must implement a destructor. */
//nsIPaperSizePS::~nsIPaperSizePS () { }

nsPaperSizePS::nsPaperSizePS() {
    mCount = COUNTOF(mList);
    mCurrent = 0;
}

PRBool
nsPaperSizePS::Find(const char *aName) {
    for (int i = mCount; i--; ) {
        if (!PL_strcasecmp(aName, mList[i].name)) {
            mCurrent = i;
            return PR_TRUE;
        }
    }
    return PR_FALSE;
}

PRBool
nsPaperSizePS::SystemDefault() {
    DO_PR_DEBUG_LOG(("nsPaperSizePS::SystemDefault () "));
    mCurrent = 0;
    if (mCount <= 0) return PR_FALSE; else return PR_TRUE;
}

nsPaperSizeCUPS::nsPaperSizeCUPS(const char* fullPrinterName,
                                 const char* printerName) {
    DO_PR_DEBUG_LOG(("nsPaperSizeCUPS::nsPaperSizeCUPS('%s', '%s')\n",
                     fullPrinterName, printerName));
    /* Don't use CUPS before we are sure we have access to the PPD */
    mUsingCups = false;
    mCount = COUNTOF(mList);
    mCurrent = 0;
    mPPD = NULL;
    mPageSizeText = NULL;
    if (!fullPrinterName || !printerName ||
        nsPSPrinterList::kTypeCUPS !=
        nsPSPrinterList::GetPrinterType(nsDependentCString(fullPrinterName))) 
        return;
    if (!mCups.IsInitialized()) { mCups.Init(); }
    if (!mCups.IsInitialized()) {
        DO_PR_DEBUG_LOG(("nsPaperSizeCUPS::nsPaperSizeCUPS: CUPS unavailable\n"));
        return;
    }
    const char* ppdFileName = mCups.mCupsGetPPD (printerName);
    if (!ppdFileName) {
        DO_PR_DEBUG_LOG(("nsPaperSizeCUPS::nsPaperSizeCUPS: "
                         "cannot get PPD file name for printer '%s'\n",
                         printerName));
        return;
    }
    mPPD = mCups.mPpdOpenFile (ppdFileName);
    if (!mPPD)  {
        DO_PR_DEBUG_LOG(("nsPaperSizeCUPS::nsPaperSizeCUPS: "
                         "cannot open PPD file '%s'\n",
                         ppdFileName));
        return;
    }
    mPageSizeOption = mCups.mPpdFindOption (mPPD, "PageSize");
    mCount = mPPD->num_sizes;
    mCups.mPpdMarkDefaults (mPPD);
    mPageSizeText = new const char* [mCount];
    if (mPageSizeText) 
        for (int i = mCount; i--; ) mPageSizeText [i] = NULL;
    mUsingCups = true;
}

nsPaperSizeCUPS::~nsPaperSizeCUPS() {
    if (mPPD) mCups.mPpdClose (mPPD);
    if (mPageSizeText) delete (mPageSizeText);
}

void
nsPaperSizeCUPS::SkipZeroSizes() {
    if (!mUsingCups) return; 
    while (mCurrent < mCount
           && (mPPD->sizes [mCurrent].width == 0.0f || 
               mPPD->sizes [mCurrent].length == 0.0f)) {
        mCurrent++;
    }
}

const char*
nsPaperSizeCUPS::GetHumanReadableName (int i) {
    const char* briefName = mPPD->sizes [i].name;
    const char* name = NULL;
    if (mPageSizeText) {
        if (!mPageSizeText [i]) {
            if (mPageSizeOption) {
                ppd_choice_t* choice
                    = mCups.mPpdFindChoice (mPageSizeOption, briefName);
                if (choice) mPageSizeText [i] = choice->text;
            }
            if (!mPageSizeText [i] 
                || PL_strcasecmp (mPageSizeText [i], "") == 0)
                mPageSizeText [i] = briefName;
        }
        name = mPageSizeText [i];
    } else {
        name = briefName;
    }
    return name;
}

PRBool
nsPaperSizeCUPS::Find(const char* aName) {
    DO_PR_DEBUG_LOG(("nsPaperSizeCUPS::Find ('%s') ", aName));
    if (!mUsingCups) return nsPaperSizePS::Find(aName);
    for (int i = mCount; i--; ) {
        const char* name = GetHumanReadableName (i);
        if (!PL_strcasecmp(aName, name)) {
            DO_PR_DEBUG_LOG
                (("found paper '%s' (%gx%gmm)\n",
                  aName,
                  round(mPPD->sizes [i].width * MM_PER_PT),
                  round(mPPD->sizes [i].length * MM_PER_PT)));
            mCurrent = i;
            return PR_TRUE;
        }
    }
    DO_PR_DEBUG_LOG(("did not find paper '%s'\n", aName));
    return PR_FALSE;
}

PRBool
nsPaperSizeCUPS::SystemDefault() {
    DO_PR_DEBUG_LOG(("nsPaperSizeCUPS::SystemDefault () "));
    if (!mUsingCups) return nsPaperSizePS::SystemDefault();
    First ();
    for (int i = mCount; i--; )
        if (mPPD->sizes [i].marked) {
            mCurrent = i;
            break;
        }
    if (mCurrent < mCount) {
        DO_PR_DEBUG_LOG
            (("selected system default paper '%s'\n",
              GetHumanReadableName (mCurrent)));
        return PR_TRUE; 
    } else {
        return PR_FALSE;
    }
}

const char*
nsPaperSizeCUPS::Name() {
    if (!mUsingCups) return nsPaperSizePS::Name();
    NS_PRECONDITION(!AtEnd(), "Invalid current item");
    return GetHumanReadableName (mCurrent);
}

float
nsPaperSizeCUPS::Width_mm() {
    if (!mUsingCups) return nsPaperSizePS::Width_mm();
    NS_PRECONDITION(!AtEnd(), "Invalid current item");
    return round(mPPD->sizes [mCurrent].width * MM_PER_PT);
}

float
nsPaperSizeCUPS::Height_mm() {
    if (!mUsingCups) return nsPaperSizePS::Height_mm();
    NS_PRECONDITION(!AtEnd(), "Invalid current item");
    return round(mPPD->sizes [mCurrent].length * MM_PER_PT);
}

PRBool 
nsPaperSizeCUPS::IsMetric() {
    if (!mUsingCups) return nsPaperSizePS::IsMetric();
    NS_PRECONDITION(!AtEnd(), "Invalid current item");
    /* Educated guess: unless sizes are integral number    */
    /* of half inches, present them to the user in metric. */
    return
        fabs(fmod(mPPD->sizes [mCurrent].width, HALF_INCH_PT)) > EPSILON ||
        fabs(fmod(mPPD->sizes [mCurrent].length, HALF_INCH_PT)) > EPSILON;
}
